from pandas_ml_common import get_pandas_object
from pandas_ml_quant.plots.utils import new_fig_ts_axis


def ta_line(df, fields, figsize=None, ax=None, **kwargs):
    data = get_pandas_object(df, fields).values

    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    ax.plot(df.index, data, **kwargs)
    return ax




from unittest import TestCase

import numpy as np
import pandas as pd
import talib

from pandas_ml_quant.indicators.multi_object import *
from pandas_ml_quant.indicators.single_object import *
from pandas_ml_quant_test.config import DF_TEST
from pandas_ml_common import Constant


class TestIndicator(TestCase):

    def test__ema(self):
        me = ta_ema(DF_TEST["Close"], 20)[-100:]
        ta = talib.EMA(DF_TEST["Close"], 20)[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__wilders(self):
        me = ta_wilders(DF_TEST["Close"], 20)[-100:]
        ta = talib.EMA(DF_TEST["Close"], 20 * 2 -1)[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__macd(self):
        my_macd = ta_macd(DF_TEST["Close"], relative=False)
        talib_macd = pd.DataFrame(talib.MACD(DF_TEST["Close"])).T

        np.testing.assert_array_almost_equal(talib_macd[-100:].values[0], my_macd[-100:].values[0])

    def test__mom(self):
        me = ta_mom(DF_TEST["Close"])[-100:]
        ta = talib.MOM(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__stddev(self):
        me = ta_stddev(DF_TEST["Close"], ddof=0)[-100:]
        ta = talib.STDDEV(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__apo(self):
        me = ta_apo(DF_TEST["Close"], relative=False)[-100:]
        ta = talib.APO(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__trix(self):
        me = ta_trix(DF_TEST["Close"])[-100:]
        ta = talib.TRIX(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__tr(self):
        me = ta_tr(DF_TEST, relative=False)[-100:]
        ta = talib.TRANGE(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__atr(self):
        me = ta_atr(DF_TEST, relative=False)[-100:]
        ta = talib.ATR(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__adx(self):
        me = ta_adx(DF_TEST, relative=False)[-100:]
        ta_pdi = talib.PLUS_DI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]
        ta_mdi = talib.MINUS_DI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        ta_pdm = talib.PLUS_DM(DF_TEST["High"], DF_TEST["Low"])[-100:]
        ta_mdm = talib.MINUS_DM(DF_TEST["High"], DF_TEST["Low"])[-100:]
        ta_dx = talib.ADX(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me["ADX"], ta_dx / 100)
        np.testing.assert_array_almost_equal(me["+DI"], ta_pdi / 100)
        np.testing.assert_array_almost_equal(me["-DI"], ta_mdi / 100)

        # I have absolutely no idea why we are off by a constant factor. but since it is a constant factor we can
        # neglect it
        np.testing.assert_array_almost_equal(me["+DM"], ta_pdm * 0.07142861354566638)
        np.testing.assert_array_almost_equal(me["-DM"], ta_mdm * 0.07142861354566638)

    def test__bbands(self):
        me = ta_bbands(DF_TEST["Close"], ddof=0)[-100:]
        u, m, l = talib.BBANDS(DF_TEST["Close"])
        u = u[-100:]
        m = m[-100:]
        l = l[-100:]

        np.testing.assert_array_almost_equal(me["mean"], m)
        np.testing.assert_array_almost_equal(me["upper"], u)
        np.testing.assert_array_almost_equal(me["lower"], l)

    def test__rsi(self):
        me = ta_rsi(DF_TEST["Close"])[-100:]
        ta = talib.RSI(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__williams_R(self):
        me = ta_williams_R(DF_TEST)[-100:]
        ta = talib.WILLR(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / -100)

    def test__ultimate(self):
        me = ta_ultimate_osc(DF_TEST)[-100:]
        ta = talib.ULTOSC(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__ppo(self):
        me = ta_ppo(DF_TEST["Close"], exponential=False)[-100:]
        ta = talib.PPO(DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test__bop(self):
        me = ta_bop(DF_TEST)[-100:]
        ta = talib.BOP(DF_TEST["Open"], DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta)

    def test__cci(self):
        me = ta_cci(DF_TEST)[-100:]
        ta = talib.CCI(DF_TEST["High"], DF_TEST["Low"], DF_TEST["Close"])[-100:]

        np.testing.assert_array_almost_equal(me, ta / 100)

    def test_cross_over(self):
        me0 = ta_cross_over(DF_TEST, "Open", "Close")
        me1 = ta_cross_over(DF_TEST[["Open", "Close"]])
        me2 = ta_cross_over(DF_TEST, "Open", DF_TEST["Close"])
        me3 = ta_cross_over(DF_TEST["Open"], DF_TEST["Close"])

        self.assertTrue(me0.iloc[-1])
        self.assertFalse(me0.iloc[-2])

        for me in [me1, me2, me3]:
            np.testing.assert_array_equal(me0.values, me.values)

    def test_cross_under(self):
        me = ta_cross_under(DF_TEST, "Open", "Close")

        print(me)
        self.assertTrue(me.iloc[-3])
        self.assertFalse(me.iloc[-2])
        self.assertFalse(me.iloc[-1])

    def test_cross_constant(self):
        o = ta_cross(DF_TEST["Close"].pct_change(), Constant(0))
        u = ta_cross(DF_TEST["Close"].pct_change(), Constant(0))

        self.assertTrue(o.iloc[-2])
        self.assertTrue(u.iloc[-5])

    def test__up_down_volatility_ratio(self):
        me = ta_up_down_volatility_ratio(DF_TEST)[-1:]

        np.testing.assert_array_almost_equal(me, np.array([[-0.067, -0.232, -0.073, -0.318, -0.319,  1.776]]), 0.001)

    def test__z_score(self):
        me = ta_zscore(DF_TEST)[-1:]

        np.testing.assert_array_almost_equal(me, np.array([[ 0.546, 0.256, 0.271, 0.457,  0.457, -0.929]]), 0.001)

    def test__ta_multi_bbands(self):
        me = ta_multi_bbands(DF_TEST["Close"])[-1:]
        print(me.values.round(3))

        np.testing.assert_array_almost_equal(
            me,
            np.array([[308.406, 309.257, 310.108, 310.959, 311.81, 312.661, 313.512, 314.363, 315.214]]),
            0.001)

    def test_ta_weekday(self):
        me = ta_week_day(DF_TEST["Close"])[-1:]

        np.testing.assert_almost_equal(0.5, me.values[0])

    def test_ta_week(self):
        me = ta_week(DF_TEST["Close"])

        np.testing.assert_almost_equal(0.9423, me.values[-1], 4)
        np.testing.assert_almost_equal(0.0192, me.loc["2017-01-03"], 4)
        np.testing.assert_almost_equal(1.0, me.loc["2017-12-29"], 4)




from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd
from scipy.stats import zscore

from pandas_ml_quant.utils import wilders_smoothing as _ws, with_column_suffix as _wcs

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_inverse(df: _PANDAS) -> _PANDAS:
    return df.apply(lambda col: col * -1 + col.min() + col.max(), raw=True)


def ta_sma(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"sma_{period}", df.rolling(period).mean())


def ta_ema(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"ema_{period}", df.ewm(span=period, adjust=False, min_periods=period-1).mean())


def ta_wilders(df: _PANDAS, period=12) -> _PANDAS:
    if isinstance(df, _pd.Series):
        res = ta_wilders(df.to_frame(), period).iloc[:, 0]
    else:
        resdf = _pd.DataFrame({}, index=df.index)
        for col in df.columns:
            s = df[col].dropna()
            res = _np.zeros(s.shape)
            _ws(s.values, period, res)
            resdf = resdf.join(_pd.DataFrame({col: res}, index=s.index))

        res = resdf

    return _wcs(f"wilders_{period}", res)


def ta_macd(df: _PANDAS, fast_period=12, slow_period=26, signal_period=9, relative=True) -> _PANDAS:
    fast = ta_ema(df, fast_period)
    slow = ta_ema(df, slow_period)
    macd = (fast / slow - 1) if relative else (fast - slow)
    signal = ta_ema(macd, signal_period)
    hist = macd - signal
    suffix = f'{fast_period},{slow_period},{signal_period}'

    for label, frame in {f"macd_{suffix}": macd, f"signal_{suffix}": signal, f"histogram_{suffix}": hist}.items():
        if isinstance(frame, _pd.DataFrame) and len(df.columns) > 1:
            frame.columns = _pd.MultiIndex.from_product([frame.columns, [label]])
        else:
            frame.name = label

    macd = macd.to_frame() if isinstance(macd, _pd.Series) else macd
    return macd.join(signal).join(hist)


def ta_mom(df: _PANDAS, period=10) -> _PANDAS:
    return _wcs(f"mom_{period}", df.diff(period))


def ta_roc(df: _PANDAS, period=10) -> _PANDAS:
    return _wcs(f"roc_{period}", df.pct_change(period))


def ta_stddev(df: _PANDAS, period=5, nbdev=1, ddof=1) -> _PANDAS:
    return _wcs(f"stddev_{period}", (df.rolling(period).std(ddof=ddof) * nbdev))


def ta_rsi(df: _PANDAS, period=14):
    returns = df.diff()

    pos = ta_wilders(returns.clip(lower=0), period)
    neg = ta_wilders(_np.abs(returns.clip(upper=0)), period)

    return _wcs(f"rsi_{period}", pos / (pos + neg), returns)


def ta_apo(df: _PANDAS, fast_period=12, slow_period=26, exponential=False, relative=True) -> _PANDAS:
    fast = ta_ema(df, fast_period) if exponential else ta_sma(df, fast_period)
    slow = ta_ema(df, slow_period) if exponential else ta_sma(df, slow_period)
    apo = (fast / slow) if relative else (fast - slow)
    return _wcs(f"apo_{fast_period},{slow_period},{int(exponential)}", apo, df)


def ta_trix(df: _PANDAS, period=30) -> _PANDAS:
    return _wcs(f"trix_{period}", ta_ema(ta_ema(ta_ema(df, period), period), period).pct_change() * 100, df)


def ta_ppo(df: _pd.DataFrame, fast_period=12, slow_period=26, exponential=True) -> _PANDAS:
    fast = ta_ema(df, period=fast_period) if exponential else ta_sma(df, period=fast_period)
    slow = ta_ema(df, period=slow_period) if exponential else ta_sma(df, period=slow_period)
    return _wcs(f"ppo_{fast_period},{slow_period},{int(exponential)}", (fast - slow) / slow, df)


def ta_zscore(df: _PANDAS, period=20, ddof=1):
    res = df.rolling(period).apply(lambda c: zscore(c, ddof=ddof)[-1])

    return _wcs(f"z_{period}", res)


def ta_decimal_year(df: _PANDAS):
    return df.index.strftime("%j").astype(float) - 1 / 366 + df.index.strftime("%Y").astype(float)


def ta_week_day(po: _PANDAS):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return (df.index.dayofweek / 6.0).to_series(index=po.index, name="dow")


def ta_week(po: _PANDAS):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return (df.index.week / 52.0).to_series(index=po.index, name="week")


def ta_ewma_covariance(df: _PANDAS, convert_to='returns', alpha=0.97):
    data = df

    if convert_to == 'returns':
        data = df.pct_change()
    if convert_to == 'log-returns':
        data = _np.log(df) - _np.log(df.shift(1))

    return data.ewm(com=alpha).cov()


def ta_up_down_volatility_ratio(df: _PANDAS, period=60, normalize=True, setof_date=False):
    if isinstance(df, _pd.DataFrame):
        return df.apply(lambda col: ta_up_down_volatility_ratio(col, period, normalize))

    returns = df.pct_change() if normalize else df
    std = _pd.DataFrame({
        "std+": returns[returns > 0].rolling(period).std(),
        "std-": returns[returns < 0].rolling(period).std()
    }, index=returns.index).fillna(method="ffill")

    ratio = (std["std+"] / std["std-"] - 1).rename("std +/-")

    # eventually we can off set the date such that we can fake one continuous data frame
    if setof_date:
        # +7 -1 binds us approximately to the same week day
        ratio.index = ratio.index - _pd.DateOffset(days=(ratio.index[-1] - ratio.index[0]).days + 7 - 1)

    return ratio


def ta_multi_bbands(s: _pd.Series, period=5, stddevs=[0.5, 1.0, 1.5, 2.0], ddof=1) -> _PANDAS:
    assert isinstance(s, _pd.Series)
    mean = s.rolling(period).mean().rename("mean")
    std = s.rolling(period).std(ddof=ddof)
    df = _pd.DataFrame({}, index=mean.index)

    for stddev in reversed(stddevs):
        df[f'lower-{stddev}'] = mean - (std * stddev)

    df["mean"] = mean

    for stddev in stddevs:
        df[f'upper-{stddev}'] = mean + (std * stddev)

    return df


def ta_bbands(df: _PANDAS, period=5, stddev=2.0, ddof=1) -> _PANDAS:
    mean = df.rolling(period).mean()
    std = df.rolling(period).std(ddof=ddof)
    most_recent = df.rolling(period).apply(lambda x: x[-1], raw=True)

    upper = mean + (std * stddev)
    lower = mean - (std * stddev)
    z_score = (most_recent - mean) / std
    quantile = (most_recent > upper).astype(int) - (most_recent < lower).astype(int)

    if isinstance(mean, _pd.Series):
        upper.name = "upper"
        mean.name = "mean"
        lower.name = "lower"
        z_score.name = "z"
        quantile.name = "quantile"
    else:
        upper.columns = _pd.MultiIndex.from_product([upper.columns, ["upper"]])
        mean.columns = _pd.MultiIndex.from_product([mean.columns, ["mean"]])
        lower.columns = _pd.MultiIndex.from_product([lower.columns, ["lower"]])
        z_score.columns = _pd.MultiIndex.from_product([z_score.columns, ["z"]])
        quantile.columns = _pd.MultiIndex.from_product([z_score.columns, ["quantile"]])

    return _pd.DataFrame(upper) \
        .join(mean) \
        .join(lower) \
        .join(z_score) \
        .join(quantile) \
        .sort_index(axis=1)



from typing import Union, Iterable, Dict, Callable, Tuple

import pandas as pd
from sortedcontainers import SortedDict


def ta_rnn(df: Union[pd.Series, pd.DataFrame],
           feature_lags: Iterable[int],
           lag_smoothing: Dict[int, Callable[[pd.Series], pd.Series]] = None,
           return_min_required_samples=False
           ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, int]]:
    df = df.to_frame()
    dff = pd.DataFrame({}, index=df.index)

    # return RNN shaped 3D arrays
    for feature in df.columns:
        feature_series = df[feature]
        smoothers = None

        # smooth out feature if requested
        if lag_smoothing is not None:
            smoothers = SortedDict({lag: smoother(feature_series.to_frame())
                                    for lag, smoother in lag_smoothing.items()})

        for lag in (feature_lags if isinstance(feature_lags, Iterable) else range(feature_lags)):
            # if smoothed values are applicable use smoothed values
            if smoothers is not None and len(smoothers) > 0 and smoothers.peekitem(0)[0] <= lag:
                feature_series = smoothers.popitem(0)[1]

            # assign the lagged (eventually smoothed) feature to the features frame
            dff[(feature, lag)] = feature_series.shift(lag)

    # fix tuple column index to actually be a multi index and fix levels
    # features need to be row, timestep, feature to fit RNN based models
    dff.columns = pd.MultiIndex.from_tuples(dff.columns).swaplevel(0, 1)

    # drop all rows which got nan now
    dff = dff.dropna()

    if return_min_required_samples:
        return dff, len(df) - len(dff)
    else:
        return dff




import numpy as np
import pandas as pd
from numba import guvectorize, float32, int32, float64, int64


@guvectorize([(float32[:], int32, float32[:]),
              (float64[:], int64, float64[:])], '(n),()->(n)')
def wilders_smoothing(arr: np.ndarray, period: int, res: np.ndarray):
    assert period > 0
    alpha = (period - 1) / period
    beta = 1 / period

    res[0:period] = np.nan
    res[period - 1] = arr[0:period].mean()
    for i in range(period, len(arr)):
        res[i] = alpha * res[i-1] + arr[i] * beta


def with_column_suffix(suffix, po, ref_po=None):
    if ref_po is None:
        ref_po = po

    if isinstance(po, pd.Series):
        if isinstance(po.name, tuple):
            return po.rename((suffix, *ref_po.name))
        else:
            return po.rename(f'{ref_po.name}_{suffix}')
    else:
        if isinstance(po.index, pd.MultiIndex):
            po.columns = pd.MultiIndex.from_tuples([(suffix, *col) for col in ref_po.columns.to_list()])
            return po
        else:
            po.columns = ref_po.columns
            return po.add_suffix(f'_{suffix}')


# return index of bucket of which the future price lies in
def index_of_bucket(value, data):
    if np.isnan(value) or np.isnan(data).any() or np.isinf(value) or np.isinf(value).any():
        return np.nan

    for i, v in enumerate(data):
        if value < v:
            return i

    return len(data)



from typing import Union as _Union
from pyts.image import GramianAngularField as _GAF
import pandas as _pd
_PANDAS = _Union[_pd.DataFrame, _pd.Series]


# FIXME
def ta_gaf(df: _pd.DataFrame,
          columm_index_level=1,
          image_size=24,
          sample_range=(-1, 1),
          method='summation',
          flatten=False,
          overlapping=False):

    if isinstance(df.columns, _pd.MultiIndex):
        # for each n'd level column
        l = columm_index_level
        columns = {c[l]: [c2 for c2 in df.columns.to_list() if c2[l] == c[l]] for c in df.columns.to_list()}
        res = _pd.DataFrame({}, index=df.index)

        for feature, timesteps in columns.items():
            dff = df[timesteps]
            dff.columns = [t[:l] + t[l+1:] for t in timesteps]
            s = ta_gaf(dff, columm_index_level, image_size, sample_range, method, flatten, overlapping)
            res[f'{feature}_gaf'] = s

        return res
    else:
        image_size = max(min(image_size, len(df.columns)), 1)
        gaf = _GAF(image_size=image_size, sample_range=sample_range, method=method,
                   overlapping=overlapping, flatten=flatten)

        return _pd.Series(list(gaf.fit_transform(df.values)), name="gaf", index=df.index)





import os

import numpy as np

from pandas_ml_common import pd


DF_TEST_MULTI = pd.read_pickle(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "spy_gld.pickle"))
DF_TEST = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "SPY.csv"), index_col='Date', parse_dates=True)
DF_DEBUG = pd.DataFrame({"Close": np.random.random(10)})



import matplotlib.dates as mdates
import matplotlib.pyplot as plt
from matplotlib import gridspec
from pandas.core.base import PandasObject


def matplot_dates(df: PandasObject):
    return mdates.date2num(df.index)


def new_fig_ts_axis(figsize=(12, 8)):
    fig = plt.figure('r-', figsize=figsize)
    grid = gridspec.GridSpec(1, 1)
    ax = fig.add_subplot(grid[0])
    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))
    plt.xticks(rotation=45)

    return fig, ax



from unittest import TestCase
from pandas_ml_quant import pd, np
from pandas_ml_quant.encoders import ta_rescale, ta_realative_candles
from pandas_ml_quant_test.config import DF_TEST


class TestRescaling(TestCase):

    def test_rescale(self):
        df = pd.DataFrame({"a": [1, 2, 3, 4, 4, 5],
                           "b": [5, 4, 3, 2, 2, 2]})

        series = ta_rescale(df["a"])
        all = ta_rescale(df)
        columns = ta_rescale(df, axis=0)
        rows = ta_rescale(df, axis=1)

        self.assertEqual(1, series.values[-1])
        self.assertListEqual([1, -0.5], all[-1:].values[-1].tolist())
        self.assertListEqual([1, -1], columns[-1:].values[-1].tolist())
        self.assertListEqual([1, -1], rows[-1:].values[-1].tolist())

    def test_relative_candles(self):
        df = DF_TEST[-2:].copy()
        relative = ta_realative_candles(df)

        np.testing.assert_array_almost_equal(np.array([0.002469,  0.002021,  0.002533, -0.002829]),
                                             relative.values[-1])



import matplotlib.dates as mdates
import pandas as pd
from mpl_finance import candlestick_ohlc

from pandas_ml_common import get_pandas_object
from pandas_ml_quant.plots.utils import new_fig_ts_axis


def ta_candlestick(self, open="Open", high="High", low="Low", close="Close", ax=None, figsize=None, **kwargs):
    df = self if isinstance(self, pd.DataFrame) else self._parent

    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    # Plot candlestick chart
    data = pd.DataFrame({
        "Date": mdates.date2num(df.index),
        "open": get_pandas_object(df, open),
        "high": get_pandas_object(df, high),
        "low": get_pandas_object(df, low),
        "close": get_pandas_object(df, close),
    })

    candlestick_ohlc(ax, data.values, width=0.6, colorup='g', colordown='r')
    return ax


from typing import List

import matplotlib.dates as mdates
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pandas.core.base import PandasObject

from pandas_ml_common import get_pandas_object
from pandas_ml_quant.encoders import ReScaler
from pandas_ml_quant.plots.utils import new_fig_ts_axis


def ta_bar(df, fields, figsize=None, ax=None, colors=None, color_map: str = 'afmhot', **kwargs):
    data = get_pandas_object(df, fields).values
    colors = get_pandas_object(df, colors)

    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    bars = ax.bar(df.index, height=data, **kwargs)
    if colors is not None:
        color_function = plt.get_cmap(color_map)
        domain = (colors.values.min(), colors.values.max()) if isinstance(colors, PandasObject) else (colors.min(), colors.max())
        r = ReScaler(domain, (0, 1))

        for i, c in enumerate(colors):
            color = color_function(r(c))
            # TODO if alpha is provided then color = (*color[:-1], alpha)
            bars[i].set_color(color)

    return ax


def ta_stacked_bar(df, columns, figsize=None, ax=None, padding=0.02, **kwargs):
    # TODO add colors ...
    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    if padding is not None:
        b, t = ax.get_ylim()

        if b == 0 and t == 1:
            b = np.inf
            t = -np.inf

        ax.set_ylim(min(df[columns].values.min(), b) * (1 - padding), max(df[columns].values.max(), t) * (1 + padding))

    bottom = None
    for column in columns:
        data = get_pandas_object(df, column)

        if bottom is not None:
            kwargs["bottom"] = bottom
            height = data - bottom
        else:
            height = data

        bottom = height if bottom is None else bottom + height
        ax.bar(mdates.date2num(df.index), height, **kwargs)

    return ax


def ta_bars(df, columns, figsize=None, ax=None, padding=0.02, **kwargs):
    # FIXME allow to pass pandas objects as columns
    columns = columns if isinstance(columns, List) else list(columns)

    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    if padding is not None:
        b, t = ax.get_ylim()

        if b == 0 and t == 1:
            b = np.inf
            t = -np.inf

        ax.set_ylim(min(df[columns].values.min(), b) * (1 - padding), max(df[columns].values.max(), t) * (1 + padding))

    width = 1 * (1-padding) / len(columns)
    for i, column in enumerate(columns):
        ax.bar(mdates.date2num(df.index) + width * i, df[column].values, width, **kwargs)

    return ax



import pandas as _pd

from pandas_ml_common import get_pandas_object as _get_pandas_object


def ta_candles_as_culb(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close", relative_close=False):
    o = _get_pandas_object(df, open)
    c = _get_pandas_object(df, close)
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    oc = _pd.concat([o, c], axis=1)
    c_1 = c.shift(1)

    # calculate close, upper_shadow, lower_shadow, body
    return _pd.DataFrame({
        "close": (c / c_1 - 1) if relative_close else c,
        "upper": (h / oc.max(axis=1) - 1),
        "lower": (oc.min(axis=1) / l - 1),
        "body": (c / o - 1)
    }, index=df.index)



from unittest import TestCase
import numpy as np
from pandas_ml_quant.utils import index_of_bucket


class TestUtils(TestCase):

    def test_index_of_bucket(self):

        a = index_of_bucket(0, np.array([1, 2, 3]))
        b = index_of_bucket(1.2, np.array([1, 2, 3]))
        c = index_of_bucket(2.1, np.array([1, 2, 3]))
        d = index_of_bucket(3, np.array([1, 2, 3]))

        self.assertEqual(0, a)
        self.assertEqual(1, b)
        self.assertEqual(2, c)
        self.assertEqual(3, d)



from collections import Callable

import tensorflow as tf
from keras import backend as K, Input
from keras.initializers import Constant
from keras.layers import Layer, Dense, Concatenate


class NormalDistributionLayer(Layer):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        pass

    def call(self, inputs, **kwargs):
        mu = Dense(1)(inputs)
        sigma = Dense(1, activation=lambda x: tf.nn.elu(x) + 1)(inputs)
        out = Concatenate()([mu, sigma])
        return out

    def compute_output_shape(self, input_shape):
        return input_shape[0], 2


class CurveFit(Layer):

    def __init__(self, parameters: int, function: Callable, initializer='uniform', **kwargs):
        super().__init__(**kwargs)
        self.parameters = parameters
        self.function = function
        self.initializer = initializer
        self.batch_size = None
        self.kernel = None

    def build(self, input_shape):
        """
        parameters: pass the number of parameters of the function you try to fit
        function: pass the function you want to fit i.e. `lambda x, args: x * sum(args)`
        """

        # set batch size
        self.batch_size = input_shape[0]

        # Create a trainable weight variable for this layer.
        self.kernel = self.add_weight(name='kernel',
                                      shape=(self.parameters,),
                                      initializer=self.initializer,
                                      trainable=True)

        # Be sure to call this at the end
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        # here we gonna invoke out function and return the result.
        # the loss function will do whatever is needed to fit this function as good as possible
        return self.function(inputs, self.kernel)

    def compute_output_shape(self, input_shape):
        return input_shape


class LinearRegressionLayer(CurveFit):

    def __init__(self):
        super().__init__(2, LinearRegressionLayer.fit)

    @staticmethod
    def fit(x, args):
        # y = k * x + d
        return args[0] * K.arange(0, x.shape[1], 1, dtype=x.dtype) + args[1]


class LPPLLayer(CurveFit):
    # original model:
    #  dt = tc - t
    #  dtPm = dt ^ m
    #  A + B * dtPm + C * dtPm * cos(w * ln(dt) - phi)
    def __init__(self):
        super().__init__(3, self.lppl, Constant(0.5))

    def get_tc(self):
        return self.get_weights()[0][0]

    def lppl(self, x, args):
        N = K.constant(int(x.shape[-1]), dtype=x.dtype)
        t = K.arange(0, int(x.shape[-1]), 1, dtype=x.dtype)

        # note that we need to get the variables to be centered around 0
        # so to correct the magnitude we offset them by constants
        # w just has a magnitude of 10s from empirical results
        # for tc we apply a factor of 5 which should be interpreted as week
        # a tc 0.5 means half a week in the future
        tc = args[0] * K.constant(5, dtype=x.dtype) + N
        m = args[1]
        w = args[2] * K.constant(10, dtype=x.dtype)

        # then we calculate the lppl with the given parameters
        dt = (tc - t)
        dtPm = K.pow(dt, m)
        dtln = K.log(dt)
        abcc = LPPLLayer.matrix_equation(x, dtPm, dtln, w, N)
        a, b, c1, c2 = (abcc[0], abcc[1], abcc[2], abcc[3])

        # LPPL = A+B(tc −t)^m +C1(tc −t)^m cos(ω ln(tc −t)) +C2(tc −t)^m sin(ω ln(tc −t))
        return a + b * dtPm + c1 * dtPm * K.cos(w * dtln) + c2 * dtPm * K.sin(w * dtln)

    @staticmethod
    def matrix_equation(x, dtPm, dtln, w, N):
        fi = dtPm
        gi = dtPm * K.cos(w * dtln)
        hi = dtPm * K.sin(w * dtln)

        fi_pow_2 = K.sum(fi * fi)
        gi_pow_2 = K.sum(gi * gi)
        hi_pow_2 = K.sum(hi * hi)

        figi = K.sum(fi * gi)
        fihi = K.sum(fi * hi)
        gihi = K.sum(gi * hi)

        # note that our price is already a log price so we should not log it one more time
        yi = x  # K.log(x)
        yifi = K.sum(yi * fi)
        yigi = K.sum(yi * gi)
        yihi = K.sum(yi * hi)

        fi = K.sum(fi)
        gi = K.sum(gi)
        hi = K.sum(hi)
        yi = K.sum(yi)

        A = K.stack([
            K.stack([N, fi, gi, hi]),
            K.stack([fi, fi_pow_2, figi, fihi]),
            K.stack([gi, figi, gi_pow_2, gihi]),
            K.stack([hi, fihi, gihi, hi_pow_2])
        ], axis=0)

        b = K.stack([yi, yifi, yigi, yihi])

        # do a classic x = (A'A)⁻¹A' b
        return tf.linalg.solve(A, K.reshape(b, (4, 1)))



from pandas_ml_common import get_pandas_object
from pandas_ml_quant.plots.utils import new_fig_ts_axis
import numpy as np


def ta_matrix(df, fields, figsize=None, ax=None, **kwargs):
    data = fields if isinstance(fields, np.ndarray) else (get_pandas_object(df, fields).ml.values.squeeze())

    if ax is None:
        fig, ax = new_fig_ts_axis(figsize)

    ax.matshow(data)
    ax.set_yticklabels([])
    ax.set_xticklabels([])

    return ax



from unittest import TestCase
from pandas_ml_quant import pd, np
from pandas_ml_quant.labels import ta_opening_gap, ta_future_bband_quantile
from pandas_ml_quant_test.config import DF_TEST


class TestDescreteLabels(TestCase):

    def test_ta_opening_gap(self):
        df = pd.DataFrame({"Open": [0, 2, 2, 2],
                           "Close":   [1, 2, 3, 0]})

        label = ta_opening_gap(df)

        self.assertListEqual([2., 0., 1.], label.dropna().tolist())

    def test_ta_future_bband_quantile(self):
        df = DF_TEST.copy()

        label1 = ta_future_bband_quantile(df["Close"]).dropna()
        label2 = ta_future_bband_quantile(df["Close"], include_mean=True).dropna()

        self.assertEqual(2.0, label1.values.max())
        self.assertEqual(3.0, label2.values.max())





from unittest import TestCase

from pandas_ml_quant import np
from pandas_ml_quant.encoders import ta_candles_as_culb
from pandas_ml_quant_test.config import DF_TEST


class TestCandleEncoder(TestCase):

    def test_multi_channel(self):
        df = DF_TEST.copy()
        club = ta_candles_as_culb(df)

        self.assertGreaterEqual(club["upper"].values.max(), 0)
        self.assertGreaterEqual(club["lower"].values.max(), 0)
        np.testing.assert_array_almost_equal(np.array([3.120900e+02, 6.402011e-05, 4.861506e-03, 9.995516e-01]),
                                             club.iloc[-1].values, 5)



from unittest import TestCase
from pandas_ml_quant import pd, np
from pandas_ml_quant.encoders import ta_gaf
from pandas_ml_quant.indicators import ta_rnn
from pandas_ml_quant_test.config import DF_TEST


class TestGAF(TestCase):

    def test_series(self):
        s = DF_TEST["Close"]
        timesteps = ta_rnn(s, [1, 2, 3])

        gaf = ta_gaf(timesteps)
        shape = gaf.ml.values.shape
        np.testing.assert_almost_equal(0.37, gaf.iloc[-1][0][0][0], 2)
        self.assertEqual((6760, 1, 3, 3), shape)

    def test_multi_channel(self):
        df = DF_TEST[["Open", "Close"]]
        timesteps = ta_rnn(df, [1, 2, 3])

        gaf = ta_gaf(timesteps)
        shape = gaf.ml.values.shape
        self.assertEqual((6760, 2, 3, 3), shape)



from unittest import TestCase
from pandas_ml_quant import pd, np
from pandas_ml_quant.encoders.one_hot import ta_one_hot_encode_discrete
from pandas_ml_quant_test.config import DF_TEST


class TestOneHotEncoder(TestCase):

    def test_one_hot_encoder_vec(self):
        df = pd.DataFrame({"a": [1, 2, 3, 4, 4, 5]})

        encoded = ta_one_hot_encode_discrete(df["a"])

        print(repr(encoded))
        self.assertListEqual(df.index.tolist(), encoded.index.tolist())
        np.testing.assert_array_equal(np.array([[1, 0, 0, 0, 0],
                                                [0, 1, 0, 0, 0],
                                                [0, 0, 1, 0, 0],
                                                [0, 0, 0, 1, 0],
                                                [0, 0, 0, 1, 0],
                                                [0, 0, 0, 0, 1]]),
                                      encoded.ml.values)

    def test_one_hot_encoder_arr(self):
        df = pd.DataFrame({"a": [1, 2, 3, 4, 4, 5],
                           "b": [1, 2, 3, 4, 4, 5]})

        encoded = ta_one_hot_encode_discrete(df[["a", "b"]])

        print(repr(encoded.ml.values))
        self.assertListEqual(df.index.tolist(), encoded.index.tolist())
        np.testing.assert_array_equal(np.array([[[1, 0, 0, 0, 0],
                                                 [1, 0, 0, 0, 0]],

                                                [[0, 1, 0, 0, 0],
                                                 [0, 1, 0, 0, 0]],

                                                [[0, 0, 1, 0, 0],
                                                 [0, 0, 1, 0, 0]],

                                                [[0, 0, 0, 1, 0],
                                                 [0, 0, 0, 1, 0]],

                                                [[0, 0, 0, 1, 0],
                                                 [0, 0, 0, 1, 0]],

                                                [[0, 0, 0, 0, 1],
                                                 [0, 0, 0, 0, 1]]]),
                                      encoded.ml.values)

    def test_chained_encoder(self):
        df = DF_TEST.copy()

        discrete = df["Close"].q.ta_future_bband_quantile(7, period=14).dropna()
        onehot = df["Close"].q.ta_future_bband_quantile(7, period=14).q.ta_one_hot_encode_discrete()

        self.assertEqual(len(discrete), len(onehot))
        self.assertEqual(1., discrete[-1:].values[0])
        self.assertListEqual([0, 1, 0], onehot.iloc[-1])


import pandas as _pd
import numpy as _np
from typing import Union as _Union
import pandas_ml_quant.indicators as _i
from pandas_ml_common import get_pandas_object as _get_pandas_object

_PANDAS = _Union[_pd.DataFrame, _pd.Series]



# FIXME
# FIXME                           OBSOLETE !!
# FIXME              functions either discrete or continuous !!!
# --------------------------------------------------------------


def _ta_future_multiband_bucket(df: _pd.Series, forecast_period=14, period=5, stddevs=[0.5, 1.0, 1.5, 2.0], ddof=1):
    buckets = _i.ta_multi_bbands(df, period, stddevs=stddevs, ddof=ddof)
    future = df.shift(-forecast_period)

    # return index of bucket of which the future price lies in
    def index_of_bucket(value, data):
        if _np.isnan(value) or _np.isnan(data).any():
            return _np.nan

        for i, v in enumerate(data):
            if value < v:
                return i

        return len(data)

    return buckets \
        .join(future) \
        .apply(lambda row: index_of_bucket(row[future.name], row[buckets.columns]), axis=1, raw=False) \
        .rename(df.name)




from typing import Union as _Union

import pandas as _pd
import numpy as _np

import pandas_ml_quant.indicators as _i
from pandas_ml_quant.utils import index_of_bucket as _index_of_bucket

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_future_crossings(df: _PANDAS, a=None, b=None, period=1, forecast=1):
    crossings = _i.ta_cross(df, a, b, period=period)

    if forecast > 1:
        crossings = _i.ta_rnn(crossings, range(1, forecast))

    return crossings.shift(-forecast)


def ta_future_bband_quantile(df: _pd.Series, forecast_period=5, period=14, stddev=2.0, ddof=1, include_mean=False):
    # we want to know if a future price is violating the current upper/lower band
    bands = _i.ta_bbands(df, period, stddev, ddof)
    bands = bands[["lower", "mean", "upper"] if include_mean else ["lower", "upper"] ]
    future = df.shift(-forecast_period)

    return bands \
        .join(future) \
        .apply(lambda row: _index_of_bucket(row[future.name], row[bands.columns]), axis=1, raw=False) \
        .rename(f"{df.name}_quantile")


def ta_opening_gap(df: _pd.DataFrame, offset=0.005, open="Open", close="Close"):
    gap = (df["Open"].shift(-1) / df["Close"]) - 1
    return gap.apply(lambda row: _np.nan if _np.isnan(row) or _np.isinf(row) else \
                                 2 if row > offset else 1 if row < -offset else 0)



from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd

from pandas_ml_common import get_pandas_object as _get_pandas_object
import pandas_ml_quant.indicators.single_object as _i

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_tr(df: _PANDAS, high="High", low="Low", close="Close", relative=False) -> _PANDAS:
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    c = _get_pandas_object(df, close).shift(1)

    if relative:
        ranges = (h / l - 1).rename("a").to_frame() \
            .join((h / c - 1).rename("b").abs()) \
            .join((l / c - 1).rename("c").abs())
    else:
        ranges = (h - l).rename("a").to_frame()\
            .join((h - c).rename("b").abs())\
            .join((l - c).rename("c").abs())

    return ranges.max(axis=1).rename("true_range")


def ta_atr(df: _PANDAS, period=14, high="High", low="Low", close="Close", relative=False, exponential='wilder') -> _PANDAS:
    if exponential is True:
        atr = _i.ta_ema(ta_tr(df, high, low, close, relative), period)
    if exponential == 'wilder':
        atr = _i.ta_wilders(ta_tr(df, high, low, close, relative), period)
    else:
        atr = _i.ta_sma(ta_tr(df, high, low, close, relative), period)

    return atr.rename(f"atr_{period}")


def ta_adx(df: _PANDAS, period=14, high="High", low="Low", close="Close", relative=True) -> _PANDAS:
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)

    temp = _pd.DataFrame({
        "up": (h / h.shift(1) - 1) if relative else (h - h.shift(1)),
        "down": (l.shift(1) / l - 1) if relative else (l.shift(1) - l)
    }, index=df.index)

    atr = ta_atr(df, period, high, low, close, relative=False)
    pdm = _i.ta_wilders(temp.apply(lambda r: r[0] if r["up"] > r["down"] and r["up"] > 0 else 0, raw=False, axis=1), period)
    ndm = _i.ta_wilders(temp.apply(lambda r: r[1] if r["down"] > r["up"] and r["down"] > 0 else 0, raw=False, axis=1), period)

    pdi = pdm / atr
    ndi = ndm / atr
    adx = _i.ta_wilders((pdi - ndi).abs() / (pdi + ndi).abs(), period)

    return _pd.DataFrame({"+DM": pdm, "-DM": ndm, "+DI": pdi, "-DI": ndi, "ADX": adx}, index=df.index)


def ta_williams_R(df: _pd.DataFrame, period=14, close="Close", high="High", low="Low") -> _pd.Series:
    temp = _get_pandas_object(df, close).to_frame()
    temp = temp.join(_get_pandas_object(df, high if high is not None else close).rolling(period).max().rename("highest_high"))
    temp = temp.join(_get_pandas_object(df, low if low is not None else close).rolling(period).min().rename("lowest_low"))
    return ((temp["highest_high"] - temp[close]) / (temp["highest_high"] - temp["lowest_low"])).rename(f"williams_R_{period}")


def ta_ultimate_osc(df: _pd.DataFrame, period1=7, period2=14, period3=28, close="Close", high="High", low="Low") -> _pd.Series:
    # BP = Close - Minimum(Low or Prior Close).
    # TR = Maximum(High or Prior Close)  -  Minimum(Low or Prior Close)
    prev_close = _get_pandas_object(df, close).shift(1)
    downs = (_get_pandas_object(df, low if low is not None else close).to_frame().join(prev_close)).min(axis=1)
    ups = (_get_pandas_object(df, high if high is not None else close).to_frame().join(prev_close)).max(axis=1)
    temp = _pd.DataFrame({
        "bp": _get_pandas_object(df, close) - downs,
        "tr": ups - downs
    }, index=df.index)

    periods = [period1, period2, period3]
    avs = []
    for period in periods:
        # Average7 = (7 - period BP Sum) / (7 - period TR Sum)
        av = temp.rolling(period).sum()
        avs.append(av["bp"] / av["tr"])

    # UO = [(4 x Average7) + (2 x Average14) + Average28] / (4 + 2 + 1)
    return ((4 * avs[0] + 2 * avs[1] + avs[2]) / 7).rename(f"ultimate_osc_{period1},{period2},{period3}")


def ta_bop(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close") -> _PANDAS:
    # (CLOSE – OPEN) / (HIGH – LOW)
    o = _get_pandas_object(df, open)
    c = _get_pandas_object(df, close)
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)

    return ((c - o) / (h - l)).rename("bop")


def ta_cci(df: _pd.DataFrame, period=14, high="High", low="Low", close="Close", alpha=0.015) -> _PANDAS:
    h = _get_pandas_object(df, high)
    l = _get_pandas_object(df, low)
    c = _get_pandas_object(df, close)

    tp = (h + l + c) / 3
    tp_sma = _i.ta_sma(tp, period)
    md = tp.rolling(period).apply(lambda x: _np.abs(x - x.mean()).sum() / period)
    return ((1 / alpha) * (tp - tp_sma) / md / 100).rename(f"cci_{period}")


def ta_cross_over(df: _pd.DataFrame, a=None, b=None, period=1) -> _PANDAS:
    # return only > 0
    return ta_cross(df, a, b, period).clip(lower=0)


def ta_cross_under(df: _pd.DataFrame, a=None, b=None, period=1) -> _PANDAS:
    # return only < 0
    return ta_cross(df, a, b, period).clip(upper=0)


def ta_cross(df: _pd.DataFrame, a=None, b=None, period=1):
    # get pandas objects for crossing
    if a is None and b is None:
        assert len(df.columns) == 2, f"ambiguous crossing of {df.columns}"
        a = df[df.columns[0]]
        b = df[df.columns[1]]
    elif b is None:
        b = _get_pandas_object(df, a)
        a = df
    elif a is None:
        b = _get_pandas_object(df, b)
        a = df
    else:
        a = _get_pandas_object(df, a)
        b = _get_pandas_object(df, b)

    # get periods
    a1 = a.shift(period)
    b1 = b.shift(period)

    # if a1 < b1 and a > b then a crosses over b
    a_over_b = ((a1 < b1) & (a > b)).astype(int)

    # if a1 > b1 and a < b then a cross under b
    a_under_b = ((a1 > b1) & (a < b)).astype(int) * -1

    return a_over_b + a_under_b



"""Augment pandas DataFrame with methods for quant analysis"""
__version__ = '0.1.1'

import sys as _sys

import pandas

import pandas_ml_quant.data.datafetching as data_fetchers
from pandas_ml_common import *
from pandas_ml_quant.df.quant import Quant

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "q", property(lambda self: Quant(self)))


if 'pandas_ml_utils' in _sys.modules:
    import pandas_ml_utils
    _log.warning(f"automatically imported pandas_ml_utils {pandas_ml_utils.__version__}")


# add data fetcher functions
for fetcher_functions in [data_fetchers]:
    for fetcher_function in dir(fetcher_functions):
        if fetcher_function.startswith("fetch_"):
            setattr(pandas, fetcher_function, getattr(fetcher_functions, fetcher_function))





import io
from contextlib import redirect_stdout

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib import gridspec
from moviepy.video.VideoClip import DataVideoClip
from moviepy.video.io.bindings import mplfig_to_npimage
from moviepy.video.io.html_tools import ipython_display
from pandas.plotting import register_matplotlib_converters

from pandas_ml_quant.plots import ta_bar, ta_stacked_bar, ta_candlestick, ta_line, ta_matrix

register_matplotlib_converters()


class TaPlot(object):

    def __init__(self, df: pd.DataFrame, figsize, rows=2, cols=1, main_height_ratio=4):
        fig = plt.figure('r-', figsize=figsize)
        grid = gridspec.GridSpec(rows, cols, height_ratios=[main_height_ratio, *[1 for _ in range(1, rows)]])
        axis = []

        for i, gs in enumerate(grid):
            ax = fig.add_subplot(gs, sharex=axis[0] if i > 0 else None)
            ax.xaxis_date()
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))

            if i < rows - 1:
                ax.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
            else:
                ax.tick_params(axis='x', labelrotation=45)

            axis.append(ax)

        plt.xticks(rotation=45)

        self.df = df
        self.x = mdates.date2num(df.index)
        self.fig = fig
        self.axis = axis
        self.grid = grid

    def candlestick(self, open="Open", high="High", low="Low", close="Close", panel=0):
        self.axis[panel] = ta_candlestick(self.df, open, high, low, close, ax=self.axis[panel])
        return self._return()

    def stacked_bar(self, columns, padding=0.02, panel=1, **kwargs):
        # todo if x axis is multilevel then stack all bars at level > 2
        self.axis[panel] = ta_stacked_bar(self.df, columns, ax=self.axis[panel], padding=padding, **kwargs)
        return self._return()

    def bars(self):
        # FIXME add side by side bars
        # todo if x axis is multilevel then stack all bars at level > 2
        pass

    def bar(self, fields="Volume", panel=1, colors=None, color_map: str = 'afmhot', **kwargs):
        self.axis[panel] = ta_bar(self.df, fields, ax=self.axis[panel], colors=colors, color_map=color_map, **kwargs)
        return self._return()

    def line(self, fields="Close", panel=0, **kwargs):
        self.axis[panel] = ta_line(self.df, fields, ax=self.axis[panel], **kwargs)
        return self._return()

    def plot_matrix_animation(self, fields, fps=2, **kwargs):
        def make_frame(index):
            fig, ax = plt.subplots(figsize=(9, 9))
            ta_matrix(self.df.loc[[index]], fields, ax=ax, **kwargs)
            frame = mplfig_to_npimage(fig)
            plt.close(fig)
            return frame

        def repr_html(clip):
            f = io.StringIO()
            with redirect_stdout(f):
                return ipython_display(clip)._repr_html_()

        animation = DataVideoClip(self.df.index, make_frame, fps=fps)

        # bind the jupyter extension to the animation and return
        setattr(DataVideoClip, 'display', ipython_display)
        setattr(DataVideoClip, '_repr_html_', repr_html)
        return animation

    def __call__(self, *args, **kwargs):
        """
        TODO i am thinking of something like df.q.ta_plot()(candlesticks=True, sma=200)
             ... more thinking needed

        :param args:
        :param kwargs:
        :return:
        """
        if "lines" in kwargs:
            self.line(kwargs.pop('lines', None), **kwargs)
        else:
            self.line()

        if "bars" in kwargs:
            self.bar(kwargs.pop('bars', None), **kwargs)
        else:
            self.bar()

    def _return(self):
        self.grid.tight_layout(self.fig)



# %matplotlib
#inline
#
#from moviepy.editor import VideoClip
#from moviepy.video.VideoClip import DataVideoClip
#from moviepy.video.io.bindings import mplfig_to_npimage
#
#samples = gaf.ml.values[:, -1][-20:]
#
#
#def make_frame(sample):
#    fig, ax_patchwork = plt.subplots(figsize=(9, 9))
#
#    ax_patchwork.matshow(sample)
#    ax_patchwork.set_title("Gramian Angular Field")
#    ax_patchwork.set_yticklabels([])
#    ax_patchwork.set_xticklabels([])
#
#    frame = mplfig_to_npimage(fig)
#    plt.close(fig)
#
#    return frame
#
#
# # animation = VideoClip(make_frame, duration=5.1)
#animation = DataVideoClip(samples, make_frame, fps=4)
#
# # if(write_gif):
# # animation.write_gif("/tmp/foo.gif",fps=1)
#
#animation.ipython_display()
#



from typing import Tuple as _Tuple

import numpy as _np
import pandas as _pd


class ReScaler(object):

    def __init__(self, domain: _Tuple[float, float], range: _Tuple[float, float]):
        self.domain = domain
        self.range = range
        self.rescale = _np.vectorize(self._rescale)

    def _interpolate(self, x: float):
        return self.range[0] * (1 - x) + self.range[1] * x

    def _uninterpolate(self, x: float):
        b = (self.domain[1] - self.domain[0]) if (self.domain[1] - self.domain[0]) != 0 else (1 / self.domain[1])
        return (x - self.domain[0]) / b

    def _rescale(self, x: float):
        return self._interpolate(self._uninterpolate(x))

    def __call__(self, *args, **kwargs):
        return self.rescale(args[0])


def ta_rescale(df: _pd.DataFrame, range=(-1, 1), axis=None):
    if axis is not None:
        return df.apply(lambda x: ta_rescale(x, range), raw=False, axis=axis, result_type='broadcast')
    else:
        rescaler = ReScaler((df.values.min(), df.values.max()), range)
        rescaled = rescaler(df)

        if len(rescaled.shape) > 1:
            return _pd.DataFrame(rescaled, columns=df.columns, index=df.index)
        else:
            return _pd.Series(rescaled, name=df.name, index=df.index)


def ta_realative_candles(df: _pd.DataFrame, open="Open", high="High", low="Low", close="Close"):
    relative = _pd.DataFrame(index=df.index)
    relative[open] = (_np.log(df[open]) - _np.log(df[close].shift(1)))
    relative[close] = (_np.log(df[close]) - _np.log(df[close].shift(1)))
    relative[high] = (_np.log(df[high]) - _np.log(df[close].shift(1)))
    relative[low] = (_np.log(df[low]) - _np.log(df[close].shift(1)))
    return relative




import pandas as pd

import pandas_ml_quant.indicators as indicators
import pandas_ml_quant.encoders as encoders
import pandas_ml_quant.labels as labels
from pandas_ml_quant.df.plot import TaPlot


class Quant(object):

    def __init__(self, df: pd.DataFrame):
        self.df = df

    def ta_plot(self, rows=2, cols=1, figsize=(18, 10)):
        return TaPlot(self.df, figsize, rows, cols)


# add wrapper to call all indicators on data frames
def wrapper(func):
    def wrapped(quant, *args, **kwargs):
        return func(quant.df, *args, **kwargs)

    return wrapped


# add indicators
for indicator_functions in [indicators, encoders, labels]:
    for indicator_function in dir(indicator_functions):
        if indicator_function.startswith("ta_"):
            setattr(Quant, indicator_function, wrapper(getattr(indicator_functions, indicator_function)))



